"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/***/
const nr_data_transtormer_1 = require("nr-data-transtormer");
const sort_direction_1 = require("./sort-direction");
const aggregate_type_1 = require("./aggregate-type");
class NrStringRender {
    /***/
    constructor() {
        /***/
        this.transformer = new nr_data_transtormer_1.NrDataTransformer(() => 'Неопределено');
    }
    /**Генерация отчета для источников данных*/
    renderReport(template, dataSources) {
        // Сгруппировать
        dataSources.push(...this.getGroupByDataSources(template, dataSources));
        template = this.removeGroupByFromTemplate(template);
        for (let dataSource of dataSources) {
            template = this.renderDataSourceRec(dataSource.name, dataSource.data, template);
            template = this.replaceParam(dataSource.param, template);
        }
        return template;
    }
    /***/
    replaceParam(param, renderCanvas) {
        for (let paramName in param) {
            if (param.hasOwnProperty(paramName)) {
                renderCanvas = this.replaceValueOnTemplate(paramName, renderCanvas, param[paramName]);
            }
        }
        return renderCanvas;
    }
    /***/
    getGroupByDataSources(template, dataSources) {
        let grouppedDataSource = [];
        for (let groupByObject of this.getGroupByObjects(template)) {
            for (let dataSource of dataSources) {
                if (dataSource.name === groupByObject.groupByCondition.dataSourceName) {
                    let groupedDataSource = this.transformer.groupByTransformer(groupByObject.groupByCondition.groupFields, dataSource.data);
                    grouppedDataSource.push({ name: groupByObject.name, data: groupedDataSource, param: {} });
                }
            }
        }
        return grouppedDataSource;
    }
    /**Генерация отчета для одного источника данных*/
    renderDataSourceRec(dataSourceName, dataSourceData, template) {
        let renderedTemplate = this.calcAggregateDataSource(dataSourceName, dataSourceData, template);
        for (let templatePart of this.getTemplatePart(dataSourceName, template)) {
            this.sortDataSourceData(dataSourceData, templatePart.sort);
            let renderedDataSource = this.renderDataSource(dataSourceName, dataSourceData, this.removeDataSourceMarker(templatePart.template, dataSourceName));
            renderedTemplate = renderedTemplate.replace(templatePart.template, renderedDataSource);
        }
        return renderedTemplate;
    }
    /***/
    calcAggregateDataSource(dataSourceName, dataSourceData, template) {
        let regExpAgg = /{{(max)?(min)?(count)?(sum)?\(([\wа-яА-ЯёЁ]+)\.([.\wа-яА-ЯёЁ]+)\)}}/g;
        let templateWithoutInnerLoop = template.replace(/{{[,.\w\n\sа-яА-ЯёЁ'\|"]+\.loop[\(\),.\w\n\sа-яА-ЯёЁ'\|"]*}}[\w\W]*{{[,.\w\n\sа-яА-ЯёЁ'\|"]+\.endloop}}/g, '');
        let expAgg;
        let templateResult = template;
        while (expAgg = regExpAgg.exec(templateWithoutInnerLoop)) {
            let dataSourceNameAgg = expAgg[5];
            if (dataSourceName === dataSourceNameAgg) {
                let aggType = (expAgg[1] || expAgg[2] || expAgg[3] || expAgg[4]);
                let aggPath = expAgg[6].split('.');
                let aggRes = this.calcAgg(dataSourceData, aggPath, aggType);
                // Заменить агрегат
                templateResult = templateResult.replace(expAgg[0], aggRes ? aggRes.toLocaleString() : '');
            }
        }
        return templateResult;
    }
    /***/
    calcAggObject(row, path, aggregateType) {
        console.log(`calcAggObject ${path.join(',')}`);
        if (path.length === 0) {
            throw new Error('Ошибка вычисления агрегата, path.length === 0');
        }
        else if (path.length === 1) {
            if (row.hasOwnProperty(path[0])) {
                return row[path[0]];
            }
            else {
                throw new Error(`В объекте ${JSON.stringify(row)} нет свойства ${path[0]}`);
            }
        }
        else {
            let childPath = [...path];
            let childProp = childPath.shift();
            if (row.hasOwnProperty(childProp)) {
                if (Array.isArray(row[childProp])) {
                    // Вычеслить агрегат
                    return this.calcAgg(row[childProp], childPath, aggregateType);
                }
                else if (row[childProp] != null && typeof row[childProp] === 'object') {
                    return this.calcAggObject(row[childProp], childPath, aggregateType);
                }
                else {
                    return null;
                }
            }
            else {
                throw new Error(`В объекте ${JSON.stringify(row)} нет свойства ${path[0]}`);
            }
        }
    }
    /***/
    calcAgg(rows, path, aggregateType) {
        console.log(`calcAgg ${path.join(',')}`);
        let aggResult = null;
        for (let row of rows) {
            let value = this.calcAggObject(row, path, aggregateType);
            if (value) {
                if (aggregateType === aggregate_type_1.AggregateType.count) {
                    if (aggResult) {
                        aggResult++;
                    }
                    else {
                        aggResult = 1;
                    }
                }
                else if (aggregateType === aggregate_type_1.AggregateType.max) {
                    if (aggResult) {
                        if (aggResult > value) {
                            aggResult = value;
                        }
                    }
                    else {
                        aggResult = value;
                    }
                }
                else if (aggregateType === aggregate_type_1.AggregateType.min) {
                    if (aggResult) {
                        if (aggResult < value) {
                            aggResult = value;
                        }
                    }
                    else {
                        aggResult = value;
                    }
                }
                else if (aggregateType === aggregate_type_1.AggregateType.sum) {
                    if (aggResult) {
                        aggResult += value;
                    }
                    else {
                        aggResult = value;
                    }
                }
            }
        }
        return aggResult;
    }
    /***/
    sortDataSourceData(dataSourceData, fields) {
        if (fields.length) {
            dataSourceData.sort((x, y) => {
                let compareResult = 0;
                fields.forEach((field, index) => {
                    let valueX = this.extractDataFromObject(x, field.name);
                    let valueY = this.extractDataFromObject(y, field.name);
                    if (valueX !== valueY || index === fields.length - 1) {
                        if (valueX > valueY) {
                            compareResult = 1;
                        }
                        else if (valueX < valueY) {
                            compareResult = -1;
                        }
                        else {
                            compareResult = 0;
                        }
                    }
                    if (field.sortDiraction === sort_direction_1.SortDirection.desc) {
                        compareResult *= -1;
                    }
                });
                return compareResult;
            });
        }
    }
    /***/
    getTemplatePart(dataSourceName, template) {
        let result = [];
        // Ищем дата сурс в шаблоне
        let regExpStr = `{{${dataSourceName}\\.loop(\\.sort\\([\\w\\n'.,\\s|]+\\))?}}[\\s\\w\\W]*?{{${dataSourceName}\\.endloop}}`;
        let regExp = new RegExp(regExpStr, 'gm');
        // Ищем шаблон
        let regExpResult = template.match(regExp);
        // Если в шаблоне есть итератор для сущности
        if (regExpResult != null && regExpResult.length > 0) {
            for (let templatePart of template.match(regExp)) {
                result.push({ template: templatePart, sort: this.getSortFields(templatePart) });
            }
        }
        return result;
    }
    /***/
    getSortFields(template) {
        let sortFields = [];
        let regExpFields = /loop\.sort\(('[\w\s\n,.'|"]+')+\)/g;
        let regExpField = /'([\w\n.]+)(\|(desc))*'[\s,]?/g;
        let sortFieldsStr;
        let sortFieldStr;
        while (sortFieldsStr = regExpFields.exec(template)) {
            while (sortFieldStr = regExpField.exec(sortFieldsStr[1])) {
                let sortDirection = sort_direction_1.SortDirection.ask;
                if (sortFieldStr.length === 4 && sortFieldStr[3] === 'desc') {
                    sortDirection = sort_direction_1.SortDirection.desc;
                }
                sortFields.push({ name: sortFieldStr[1], sortDiraction: sortDirection });
            }
        }
        return sortFields;
    }
    /**Убираем из шаблона указатели на источник данных*/
    removeDataSourceMarker(templatePart, dataSourceName) {
        let iterateItemClear = templatePart.replace(new RegExp(`^{{${dataSourceName}\\.loop(\\.sort\\([\\w\\n'.,\\s|]+\\))?}}`, 'g'), '');
        iterateItemClear = iterateItemClear.replace(new RegExp(`{{${dataSourceName}\\.endloop}}$`, 'g'), '');
        return iterateItemClear;
    }
    /**Из шаблона генерим пачку интерполированных шаблонов*/
    renderDataSource(dataSourceName, dataSourceData, template) {
        let renderedTemplates = '';
        for (let obj of dataSourceData) {
            let renderRow = template;
            // Провести интерполяцию
            for (let prop in obj) {
                if (obj.hasOwnProperty(prop)) {
                    if (Array.isArray(obj[prop])) {
                        // Рекурсия дальше
                        renderRow = this.renderDataSourceRec(prop, obj[prop], renderRow);
                    }
                    else {
                        let allProps = [prop];
                        if (obj[prop] != null && typeof obj[prop] === 'object') {
                            allProps = this.getLeafPath(obj[prop], prop);
                        }
                        for (let propItem of allProps) {
                            let valuetToReplace = this.extractDataFromObject(obj, propItem);
                            if (valuetToReplace == null) {
                                valuetToReplace = '';
                            }
                            // Регулярное выражение поиска полей интерполяции
                            let fieldName = `${dataSourceName}\\.${propItem}`;
                            renderRow = this.replaceValueOnTemplate(fieldName, renderRow, valuetToReplace);
                        }
                    }
                }
            }
            // Удалить все не интерполированные поля
            renderRow = renderRow.replace(/{{[\s\w\W]+?}}/g, '');
            renderedTemplates += renderRow;
        }
        return renderedTemplates;
    }
    /***/
    replaceValueOnTemplate(fieldName, template, valueToReplace) {
        let refExpValue = new RegExp(`{{${fieldName}(\\|date)?(\\|number)?(\\|currency)?}}`, 'g');
        let dataSourceValue = template.match(refExpValue);
        if (dataSourceValue != null) {
            // Меняем в шаблоне
            for (let iterateItemValue of dataSourceValue) {
                let valuetToReplaceFormat = valueToReplace;
                if (iterateItemValue.includes('|date')) {
                    let dt = new Date(valueToReplace);
                    valuetToReplaceFormat = this.formatApp(dt);
                }
                else if (iterateItemValue.includes('|number')) {
                    let num = +valueToReplace;
                    let language = navigator.language;
                    valuetToReplaceFormat = num.toLocaleString(language);
                }
                else if (iterateItemValue.includes('|currency')) {
                    let num = +valueToReplace;
                    let language = navigator.language;
                    valuetToReplaceFormat = num.toLocaleString(language, { 'minimumFractionDigits': 2 });
                }
                template = template.replace(iterateItemValue, valuetToReplaceFormat);
            }
        }
        return template;
    }
    /***/
    formatApp(date) {
        let mm = date.getMonth() + 1; // getMonth() is zero-based
        let dd = date.getDate();
        return [
            (dd > 9 ? '' : '0') + dd,
            (mm > 9 ? '' : '0') + mm,
            date.getFullYear()
        ].join('.');
    }
    /***/
    getGroupByObjects(template) {
        let regExpGroup = /{{([\n\wа-яёА-ЯЁ]+)\s*=\s*groupBy[ ]*\(([\n\w\s.,а-яёА-ЯЁ'"[\]]+)\)}}/g;
        let result = [];
        let regExpResult;
        while (regExpResult = regExpGroup.exec(template)) {
            let groupByCondition = this.getGroupByCondition(regExpResult[2]);
            result.push({ name: regExpResult[1], groupByCondition, fullCondition: regExpResult[0] });
        }
        return result;
    }
    /***/
    removeGroupByFromTemplate(template) {
        for (let groupByObject of this.getGroupByObjects(template)) {
            template = template.replace(groupByObject.fullCondition, '');
        }
        return template;
    }
    /***/
    getGroupByCondition(groupByCondition) {
        let regExp = /^([\n\w\sа-яёА-ЯЁ]+),\[([\n\w\s.,а-яёА-ЯЁ'"]+)\]$/g;
        let result = [];
        let regExpResult;
        while (regExpResult = regExp.exec(groupByCondition)) {
            result.push([regExpResult[1], regExpResult[2]]);
        }
        if (result.length !== 1) {
            throw new InvalidGroupObjectParams('');
        }
        return { dataSourceName: result[0][0], groupFields: this.getGroupByFields(result[0][1]) };
    }
    /***/
    getGroupByFields(groupFields) {
        let groupField = groupFields.split(',').map(x => x.trim().replace(/['"]+/g, ''));
        if (groupField.length === 0) {
            throw new InvalidGroupObjectParams('');
        }
        return groupField;
    }
    /***/
    getLeafPath(obj, selfKey) {
        if (obj != null && typeof obj === 'object') {
            let result = [];
            for (let key in obj) {
                if (obj.hasOwnProperty(key)) {
                    result.push(...this.getLeafPath(obj[key], key)
                        .map(x => `${selfKey == null ? '' : `${selfKey}.`}${x}`));
                }
            }
            return result;
        }
        else {
            return [selfKey];
        }
    }
    /***/
    extractDataFromObject(obj, path) {
        let regExp = /^(\w*)\./gm;
        let result = regExp.exec(path);
        if (result) {
            let newPath = path.replace(regExp, '');
            if (obj.hasOwnProperty(result[1])) {
                return this.extractDataFromObject(obj[result[1]], newPath);
            }
            else {
                throw Error('Не найдено свойство ' + result[1]);
            }
        }
        else {
            if (obj.hasOwnProperty(path)) {
                return obj[path];
            }
            else {
                throw Error('Не найдено свойство ' + path);
            }
        }
    }
}
exports.NrStringRender = NrStringRender;
